/*
 * Copyright 2024, gRPC Authors All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import GRPCCodeGen
import GRPCProtobufCodeGen
import SwiftProtobuf
import SwiftProtobufPluginLibrary
import Testing

@Suite
struct ProtobufCodeGenParserTests {
  @Suite("Self-contained service (test-service.proto)")
  struct TestService: UsesDescriptorSet {
    static let descriptorSetName = "test-service"
    static let fileDescriptorName = "test-service"

    @available(gRPCSwiftProtobuf 1.0, *)
    var codeGen: CodeGenerationRequest {
      get throws {
        let descriptor = try Self.fileDescriptor
        return try parseDescriptor(descriptor)
      }
    }

    @Test("Filename")
    @available(gRPCSwiftProtobuf 1.0, *)
    func fileName() throws {
      #expect(try self.codeGen.fileName == "test-service.proto")
    }

    @Test("Leading trivia")
    @available(gRPCSwiftProtobuf 1.0, *)
    func leadingTrivia() throws {
      let expected = """
        /// Leading trivia.

        // DO NOT EDIT.
        // swift-format-ignore-file
        // swiftlint:disable all
        //
        // Generated by the gRPC Swift generator plugin for the protocol buffer compiler.
        // Source: test-service.proto
        //
        // For information on using the generated types, please see the documentation:
        //   https://github.com/grpc/grpc-swift

        """

      #expect(try self.codeGen.leadingTrivia == expected)
    }

    @Test("Dependencies")
    @available(gRPCSwiftProtobuf 1.0, *)
    func dependencies() throws {
      let expected: [GRPCCodeGen.Dependency] = [
        .init(module: "GRPCProtobuf", accessLevel: .internal),  // Always an internal import
        .init(module: "SwiftProtobuf", accessLevel: .internal),
      ]
      #expect(try self.codeGen.dependencies == expected)
    }

    @Suite("Service")
    struct Service {
      @available(gRPCSwiftProtobuf 1.0, *)
      var service: GRPCCodeGen.ServiceDescriptor {
        get throws {
          let request = try parseDescriptor(try TestService.fileDescriptor)
          try #require(request.services.count == 1)
          return try #require(request.services.first)
        }
      }

      @Test("Name")
      @available(gRPCSwiftProtobuf 1.0, *)
      func name() throws {
        #expect(try self.service.name.identifyingName == "test.TestService")
      }

      @Suite("Methods")
      struct Methods {
        @available(gRPCSwiftProtobuf 1.0, *)
        var service: GRPCCodeGen.ServiceDescriptor {
          get throws {
            let request = try parseDescriptor(try TestService.fileDescriptor)
            try #require(request.services.count == 1)
            return try #require(request.services.first)
          }
        }

        @available(gRPCSwiftProtobuf 1.0, *)
        var unary: GRPCCodeGen.MethodDescriptor {
          get throws { try self.service.methods[0] }
        }
        @available(gRPCSwiftProtobuf 1.0, *)
        var clientStreaming: GRPCCodeGen.MethodDescriptor {
          get throws { try self.service.methods[1] }
        }
        @available(gRPCSwiftProtobuf 1.0, *)
        var serverStreaming: GRPCCodeGen.MethodDescriptor {
          get throws { try self.service.methods[2] }
        }
        @available(gRPCSwiftProtobuf 1.0, *)
        var bidiStreaming: GRPCCodeGen.MethodDescriptor {
          get throws { try self.service.methods[3] }
        }

        @Test("Documentation")
        @available(gRPCSwiftProtobuf 1.0, *)
        func documentation() throws {
          #expect(try self.unary.documentation == "/// Unary docs.\n")
          #expect(try self.clientStreaming.documentation == "/// Client streaming docs.\n")
          #expect(try self.serverStreaming.documentation == "/// Server streaming docs.\n")
          #expect(try self.bidiStreaming.documentation == "/// Bidirectional streaming docs.\n")
        }

        @Test("Name")
        @available(gRPCSwiftProtobuf 1.0, *)
        func name() throws {
          try #expect(self.unary.name.identifyingName == "Unary")
          try #expect(self.clientStreaming.name.identifyingName == "ClientStreaming")
          try #expect(self.serverStreaming.name.identifyingName == "ServerStreaming")
          try #expect(self.bidiStreaming.name.identifyingName == "BidirectionalStreaming")
        }

        @Test("Input")
        @available(gRPCSwiftProtobuf 1.0, *)
        func input() throws {
          #expect(try self.unary.inputType == "Test_TestInput")
          #expect(try !self.unary.isInputStreaming)

          #expect(try self.clientStreaming.inputType == "Test_TestInput")
          #expect(try self.clientStreaming.isInputStreaming)

          #expect(try self.serverStreaming.inputType == "Test_TestInput")
          #expect(try !self.serverStreaming.isInputStreaming)

          #expect(try self.bidiStreaming.inputType == "Test_TestInput")
          #expect(try self.bidiStreaming.isInputStreaming)
        }

        @Test("Output")
        @available(gRPCSwiftProtobuf 1.0, *)
        func output() throws {
          #expect(try self.unary.outputType == "Test_TestOutput")
          #expect(try !self.unary.isOutputStreaming)

          #expect(try self.clientStreaming.outputType == "Test_TestOutput")
          #expect(try !self.clientStreaming.isOutputStreaming)

          #expect(try self.serverStreaming.outputType == "Test_TestOutput")
          #expect(try self.serverStreaming.isOutputStreaming)

          #expect(try self.bidiStreaming.outputType == "Test_TestOutput")
          #expect(try self.bidiStreaming.isOutputStreaming)
        }
      }
    }
  }

  @Suite("Multi-service file (foo-service.proto)")
  struct FooService: UsesDescriptorSet {
    static let descriptorSetName = "foo-service"
    static let fileDescriptorName = "foo-service"

    @available(gRPCSwiftProtobuf 1.0, *)
    var codeGen: CodeGenerationRequest {
      get throws {
        let descriptor = try Self.fileDescriptor
        return try parseDescriptor(descriptor)
      }
    }

    @Test("Name")
    @available(gRPCSwiftProtobuf 1.0, *)
    func name() throws {
      #expect(try self.codeGen.fileName == "foo-service.proto")
    }

    @Test("Dependencies")
    @available(gRPCSwiftProtobuf 1.0, *)
    func dependencies() throws {
      let expected: [GRPCCodeGen.Dependency] = [
        .init(module: "GRPCProtobuf", accessLevel: .internal)  // Always an internal import
      ]
      #expect(try self.codeGen.dependencies == expected)
    }

    @Test("Service1")
    @available(gRPCSwiftProtobuf 1.0, *)
    func service1() throws {
      let service = try self.codeGen.services[0]
      #expect(service.name.identifyingName == "foo.FooService1")
      #expect(service.methods.count == 1)
    }

    @Test("Service1.Method")
    @available(gRPCSwiftProtobuf 1.0, *)
    func service1Method() throws {
      let method = try self.codeGen.services[0].methods[0]
      #expect(method.name.identifyingName == "Foo")
      #expect(method.inputType == "Foo_FooInput")
      #expect(method.outputType == "Foo_FooOutput")
    }

    @Test("Service2")
    @available(gRPCSwiftProtobuf 1.0, *)
    func service2() throws {
      let service = try self.codeGen.services[1]
      #expect(service.name.identifyingName == "foo.FooService2")
      #expect(service.methods.count == 1)
    }

    @Test("Service2.Method")
    @available(gRPCSwiftProtobuf 1.0, *)
    func service2Method() throws {
      let method = try self.codeGen.services[1].methods[0]
      #expect(method.name.identifyingName == "Foo")
      #expect(method.inputType == "Foo_FooInput")
      #expect(method.outputType == "Foo_FooOutput")
    }
  }

  @Suite("Service with no package (bar-service.proto)")
  struct BarService: UsesDescriptorSet {
    static var descriptorSetName: String { "bar-service" }
    static var fileDescriptorName: String { "bar-service" }

    @Test("Service name")
    @available(gRPCSwiftProtobuf 1.0, *)
    func serviceName() throws {
      let descriptor = try Self.fileDescriptor
      let codeGen = try parseDescriptor(descriptor)
      let service = try #require(codeGen.services.first)

      #expect(service.name.identifyingName == "BarService")
    }
  }

  @Suite("Service using 'well-known types' (wkt-service.proto)")
  struct WKTService: UsesDescriptorSet {
    static let descriptorSetName = "wkt-service"
    static let fileDescriptorName = "wkt-service"

    @Test("Dependencies")
    @available(gRPCSwiftProtobuf 1.0, *)
    func dependencies() throws {
      let descriptor = try Self.fileDescriptor
      let codeGen = try parseDescriptor(descriptor)

      let expected: [Dependency] = [
        Dependency(module: "GRPCProtobuf", accessLevel: .internal),
        Dependency(module: "SwiftProtobuf", accessLevel: .internal),
      ]
      #expect(codeGen.dependencies == expected)
    }
  }
}
